import json
import os.path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nomad.datamodel.data import (
        ArchiveSection,
    )
    from nomad.datamodel.datamodel import (
        EntryArchive,
    )


def get_reference(upload_id: str, entry_id: str) -> str:
    return f'../uploads/{upload_id}/archive/{entry_id}#data'


def get_entry_id_from_file_name(file_name: str, archive: 'EntryArchive') -> str:
    from nomad.utils import hash

    return hash(archive.metadata.upload_id, file_name)


def create_archive(
    entity: 'ArchiveSection',
    archive: 'EntryArchive',
    file_name: str,
) -> str:
  
    from nomad.datamodel.context import ClientContext
    
    entity_entry = entity.m_to_dict(with_root_def=True)
    if isinstance(archive.m_context, ClientContext):
        with open(file_name, 'w') as outfile:
            json.dump({'data': entity_entry}, outfile, indent=4)
        return os.path.abspath(file_name)
    if not archive.m_context.raw_path_exists(file_name):
        with archive.m_context.raw_file(file_name, 'w') as outfile:
            json.dump({'data': entity_entry}, outfile)
        archive.m_context.process_updated_raw_file(file_name)
    return get_reference(
        archive.metadata.upload_id, get_entry_id_from_file_name(file_name, archive)
    )



from nomad.config.models.plugins import ExampleUploadEntryPoint

catalysis = ExampleUploadEntryPoint(
    title='Heterogeneous Catalysis Example',
    category='FAIRmat examples',
    description=(
        'This example contains data entries for a catalyst sample and a catalytic '
        'reaction to demonstrate the new schemas from the '
        '[nomad-catalysis plugin](https://fairmat-nfdi.github.io/nomad-catalysis-plugin/index.html).'
    ),
    resources=['example_uploads/template_example/*'],
)



from nomad.config.models.plugins import ParserEntryPoint


class CatalysisParserEntryPoint(ParserEntryPoint):
    def load(self):
        from nomad_catalysis.parsers.catalysis_parsers import CatalysisParser # noqa: PLC0415, I001

        return CatalysisParser(**self.dict())


catalysis = CatalysisParserEntryPoint(
    name='CatalysisParser',
    description='A parser for catalysis data.',
    mainfile_name_re=r'.*CatalyticReaction\.(xlsx|csv)',
)


class CatalysisCollectionParserEntryPoint(ParserEntryPoint):
    def load(self):
        from nomad_catalysis.parsers.catalysis_parsers import CatalysisCollectionParser # noqa: PLC0415, I001

        return CatalysisCollectionParser(**self.dict())


catalysis_collection = CatalysisCollectionParserEntryPoint(
    name='CatalysisCollectionParser',
    description='A parser for a collection of catalysis entries.',
    mainfile_name_re=r'.*Cataly.+Collection\.(xlsx|csv)',
)



def test_importing_app():
    # this will raise an exception if pydantic model validation fails for th app
    from nomad_catalysis.apps import catalysis

    assert catalysis.app.label == 'Heterogeneous Catalysis'



chemical_data = {
    'carbon monoxide': {
        'pub_chem_id': 281,
        'iupac_name': 'carbon monoxide',
        'molecular_formula': 'CO',
        'molar_mass': 28.01,
        'inchi': 'InChI=1S/CO/c1-2',
        'inchi_key': 'UGFAIRIUMAVXCW-UHFFFAOYSA-N',
        'molecular_mass': 27.994914619,
        'monoisotopic_mass': 27.994914619,
        'smile': '[C-]#[O+]',
        'canonical_smile': '[C-]#[O+]',
        'cas_number': '630-08-0',
    },
    'CO': 'carbon monoxide',  # Reference to 'CO'
    'co': 'carbon monoxide',  # Reference to 'carbon monoxide'
    'carbon dioxide': {
        'pub_chem_id': 280,
        'iupac_name': 'carbon dioxide',
        'molecular_formula': 'CO2',
        'molecular_mass': 43.989829239,
        'molar_mass': 44.009,
        'monoisotopic_mass': 43.989829239,
        'inchi': 'InChI=1S/CO2/c2-1-3',
        'inchi_key': 'CURLTUGMZLYLDI-UHFFFAOYSA-N',
        'smile': 'C(=O)=O',
        'canonical_smile': 'C(=O)=O',
        'cas_number': '124-38-9',
    },
    'CO2': 'carbon dioxide',  # Reference to 'carbon dioxide'
    'co2': 'carbon dioxide',  # Reference to 'carbon dioxide'
    'propanoic acid': {
        'pub_chem_id': 1032,
        'iupac_name': 'propanoic acid',
        'molecular_formula': 'C3H6O2',
        'molar_mass': 74.08,
    },
    'propionic acid': 'propanoic acid',  # Reference to 'propanoic acid'
    'ammonia': {
        'pub_chem_id': 222,
        'iupac_name': 'ammonia',  # actually 'azane'
        'molecular_formula': 'H3N',
        'molecular_mass': 17.0265491,
        'molar_mass': 17.031,
        'monoisotopic_mass': 17.0265491,
        'inchi': 'InChI=1S/H3N/h1H3',
        'inchi_key': 'QGZKDVFQNNGYKY-UHFFFAOYSA-N',
        'smile': 'N',
        'canonical_smile': 'N',
        'cas_number': '7664-41-7',
    },
    'NH3': 'ammonia',  # Reference to 'ammonia'
    'nh3': 'ammonia',  # Reference to 'ammonia'
    'Ammonia': 'ammonia',  # Reference to 'ammonia'
    'molecular hydrogen': {
        'pub_chem_id': 783,
        'iupac_name': 'molecular hydrogen',
        'molecular_formula': 'H2',
        'molecular_mass': 2.0156500638,
        'molar_mass': 2.016,
        'monoisotopic_mass': 2.0156500638,
        'inchi': 'InChI=1S/H2/h1H',
        'inchi_key': 'UFHFLCQGNIYNRP-UHFFFAOYSA-N',
        'smile': '[HH]',
        'canonical_smile': '[HH]',
        'cas_number': '1333-74-0',
    },
    'H2': 'molecular hydrogen',  # Reference to 'molecular hydrogen'
    'h2': 'molecular hydrogen',  # Reference to 'molecular hydrogen'
    'hydrogen': 'molecular hydrogen',  # Reference to 'molecular hydrogen'
    'water': {
        'pub_chem_id': 962,
        'iupac_name': 'water',  # actually 'oxidane'
        'molecular_formula': 'H2O',
        'molecular_mass': 18.010564683,
        'molar_mass': 18.015,
        'monoisotopic_mass': 18.010564683,
        'inchi': 'InChI=1S/H2O/h1H2',
        'inchi_key': 'XLYOFNOQVPJJNP-UHFFFAOYSA-N',
        'smile': 'O',
        'canonical_smile': 'O',
        'cas_number': '7732-18-5',
    },
    'H2O': 'water',  # Reference to 'water'
    'h2o': 'water',  # Reference to 'water'
    'argon': {
        'pub_chem_id': 23968,
        'iupac_name': 'argon',
        'molecular_formula': 'Ar',
        'molecular_mass': 39.96238312,
        'molar_mass': 39.9,
        'monoisotopic_mass': 39.96238312,
        'inchi': 'InChI=1S/Ar',
        'inchi_key': 'XKRFYHLGVUSROY-UHFFFAOYSA-N',
        'smile': '[Ar]',
        'canonical_smile': '[Ar]',
        'cas_number': '7440-37-1',
    },
    'Ar': 'argon',  # Reference to 'argon'
    'ar': 'argon',  # Reference to 'argon'
    'helium': {
        'pub_chem_id': 23987,
        'iupac_name': 'helium',
        'molecular_formula': 'He',
        'molecular_mass': 4.002603254,
        'molar_mass': 4.0026,
        'monoisotopic_mass': 4.002603254,
        'inchi': 'InChI=1S/He',
        'inchi_key': 'SWQJXJOGLNCZEY-UHFFFAOYSA-N',
        'smile': '[He]',
        'canonical_smile': '[He]',
        'cas_number': '7440-59-7',
    },
    'He': 'helium',  # Reference to 'helium'
    'he': 'helium',  # Reference to 'helium'
    'molecular nitrogen': {
        'pub_chem_id': 947,
        'iupac_name': 'molecular nitrogen',
        'molecular_formula': 'N2',
        'molecular_mass': 28.006148008,
        'molar_mass': 28.014,
        'monoisotopic_mass': 28.006148008,
        'inchi': 'InChI=1S/N2/c1-2',
        'inchi_key': 'IJGRMHOSHXDMSA-UHFFFAOYSA-N',
        'smile': 'N#N',
        'canonical_smile': 'N#N',
        'cas_number': '7727-37-9',
    },
    'nitrogen': 'molecular nitrogen',  # Reference to 'molecular nitrogen'
    'N2': 'molecular nitrogen',  # Reference to 'molecular nitrogen'
    'n2': 'molecular nitrogen',  # Reference to 'molecular nitrogen'
    'molecular oxygen': {
        'pub_chem_id': 977,
        'iupac_name': 'molecular oxygen',
        'molecular_formula': 'O2',
        'molecular_mass': 31.989829239,
        'molar_mass': 31.999,
        'monoisotopic_mass': 31.989829239,
        'inchi': 'InChI=1S/O2/c1-2',
        'inchi_key': 'MYMOFIZGZYHOMD-UHFFFAOYSA-N',
        'smile': 'O=O',
        'canonical_smile': 'O=O',
        'cas_number': '7782-44-7',
    },
    'oxygen': 'molecular oxygen',  # Reference to 'molecular oxygen'
    'O2': 'molecular oxygen',  # Reference to 'molecular oxygen'
    'o2': 'molecular oxygen',  # Reference to 'molecular oxygen'
    'methane': {
        'pub_chem_id': 297,
        'iupac_name': 'methane',
        'molecular_formula': 'CH4',
        'molecular_mass': 16.0313001276,
        'molar_mass': 16.043,
        'monoisotopic_mass': 16.0313001276,
        'inchi': 'InChI=1S/CH4/h1H4',
        'inchi_key': 'VNWKTOKETHGBQD-UHFFFAOYSA-N',
        'smile': 'C',
        'canonical_smile': 'C',
        'cas_number': '74-82-8',
    },
    'CH4': 'methane',  # Reference to 'methane'
    'ch4': 'methane',  # Reference to 'methane'
    'ethane': {
        'pub_chem_id': 6324,
        'iupac_name': 'ethane',
        'molecular_formula': 'C2H6',
        'molar_mass': 30.07,
        'inchi': 'InChI=1S/C2H6/c1-2/h1-2H3',
        'inchi_key': 'OTMSDBZUPAUEDD-UHFFFAOYSA-N',
        'cas_number': '74-84-0',
    },
    'C2H6': 'ethane',  # Reference to 'ethane'
    'c2h6': 'ethane',  # Reference to 'ethane'
    'ethene': {
        'pub_chem_id': 6325,
        'iupac_name': 'ethene',
        'molecular_formula': 'C2H4',
        'molar_mass': 28.05,
        'inchi': 'InChI=1S/C2H4/c1-2/h1-2H2',
        'inchi_key': 'VGGSQFUCUMXWEO-UHFFFAOYSA-N',
        'cas_number': '74-85-1',
    },
    'ethylene': 'ethene',  # Reference to 'ethene'
    'C2H4': 'ethene',  # Reference to 'ethene'
    'c2h4': 'ethene',  # Reference to 'ethene'
    'ethyne': {
        'pub_chem_id': 6326,
        'iupac_name': 'ethyne',
        'molecular_formula': 'C2H2',
        'molar_mass': 26.04,
        'inchi': 'InChI=1S/C2H2/c1-2/h1-2H',
        'inchi_key': 'HSFWRNGVRCDJHI-UHFFFAOYSA-N',
        'cas_number': '74-86-2',
    },
    'acetylene': 'ethyne',  # Reference to 'ethyne'
    'C2H2': 'ethyne',  # Reference to 'ethyne'
    'c2h2': 'ethyne',  # Reference to 'ethyne'
    'ethin': 'ethyne',  # Reference to 'ethyne'
    'acetic acid': {
        'pub_chem_id': 176,
        'iupac_name': 'acetic acid',
        'molecular_formula': 'C2H4O2',
        'molecular_mass': 60.021129366,
        'molar_mass': 60.05,
        'monoisotopic_mass': 60.021129366,
        'inchi': 'InChI=1S/C2H4O2/c1-2(3)4/h1H3,(H,3,4)',
        'inchi_key': 'QTBSBXVTEAMEQO-UHFFFAOYSA-N',
        'smile': 'CC(=O)O',
        'canonical_smile': 'CC(=O)O',
        'cas_number': '64-19-7',
    },
    'aceticacid': 'acetic acid',  # Reference to 'acetic acid'
    'propane': {
        'pub_chem_id': 6334,
        'iupac_name': 'propane',
        'molecular_formula': 'C3H8',
        'molar_mass': 44.10,
        'inchi': 'InChI=1S/C3H8/c1-3-2/h3H2,1-2H3',
        'inchi_key': 'ATUOYWHBWRKTHZ-UHFFFAOYSA-N',
        'cas_number': '74-98-6',
    },
    'C3H8': 'propane',  # Reference to 'propane'
    'propene': {
        'pub_chem_id': 8252,
        'iupac_name': 'prop-1-ene',
        'molecular_formula': 'C3H6',
        'molar_mass': 42.08,
        'inchi': 'InChI=1S/C3H6/c1-3-2/h3H,1H2,2H3',
        'cas_number': '115-07-1',
    },
    'propylene': 'propene',  # Reference to 'propene'
    'C3H6': 'propene',  # Reference to 'propene'
    'c3h6': 'propene',  # Reference to 'propene'
    'propyne': {
        'pub_chem_id': 6335,
        'iupac_name': 'propyne',
        'molecular_formula': 'C3H4',
        'molar_mass': 40.06,
        'inchi': 'InChI=1S/C3H4/c1-3-2/h1H,3H2',
    },
    'propine': 'propyne',  # Reference to 'propyne'
    'propin': 'propyne',  # Reference to 'propyne'
    'C3H4': 'propyne',  # Reference to 'propyne'
    'c3h4': 'propyne',  # Reference to 'propyne'
    'prop-2-enoic acid': {
        'pub_chem_id': 6581,
        'iupac_name': 'prop-2-enoic acid',
        'molecular_formula': 'C3H4O2',
        'molar_mass': 72.06,
        'inchi': 'InChI=1S/C3H4O2/c1-2-3(4)5/h2H,1H2,(H,4,5)',
        'inchi_key': 'NIXOWILDQLNWCW-UHFFFAOYSA-N',
        'cas_number': '79-10-7',
    },
    'acrylic acid': 'prop-2-enoic acid',
    'acetaldehyde': {
        'pub_chem_id': 177,
        'iupac_name': 'acetaldehyde',
        'molecular_formula': 'C2H4O',
        'molecular_mass': 44.026214747,
        'molar_mass': 44.05,
        'monoisotopic_mass': 44.026214747,
        'inchi': 'InChI=1S/C2H4O/c1-2-3/h2H,1H3',
        'inchi_key': 'IKHGUXGNUITLKF-UHFFFAOYSA-N',
        'smile': 'CC=O',
        'canonical_smile': 'CC=O',
        'cas_number': '75-07-0',
    },
    'butane': {
        'pub_chem_id': 7843,
        'iupac_name': 'butane',
        'molecular_formula': 'C4H10',
        'molecular_mass': 58.078250319,
        'molar_mass': 58.12,
        'monoisotopic_mass': 58.078250319,
        'inchi': 'InChI=1S/C4H10/c1-3-4-2/h3-4H2,1-2H3',
        'inchi_key': 'IJDNQMDRQITEOD-UHFFFAOYSA-N',
        'smile': 'CCCC',
        'canonical_smile': 'CCCC',
        'cas_number': '106-97-8',
    },
    'n-butane': 'butane',  # Reference to 'butane'
    'n-Butane': 'butane',  # Reference to 'butane'
    'nbutane': 'butane',  # Reference to 'butane'
    '2-methylpropane': {
        'pub_chem_id': 6360,
        'iupac_name': '2-methylpropane',
        'molecular_formula': 'C4H10',
        'molecular_mass': 58.078250319,
        'molar_mass': 58.12,
        'monoisotopic_mass': 58.078250319,
        'inchi': 'InChI=1S/C4H10/c1-4(2)3/h4H,1-3H3',
        'inchi_key': 'NNPPMTNAJDCUHE-UHFFFAOYSA-N',
        'smile': 'CC(C)C',
        'canonical_smile': 'CC(C)C',
    },
    'isobutane': '2-methylpropane',
    'Isobutane': '2-methylpropane',
    'ibutane': '2-methylpropane',
    'butan-1-ol': {
        'pub_chem_id': 263,
        'iupac_name': 'butan-1-ol',
        'molecular_formula': 'C4H10O',
        'molecular_mass': 74.073164938,
        'molar_mass': 74.12,
        'monoisotopic_mass': 74.073164938,
        'inchi': 'InChI=1S/C4H10O/c1-2-3-4-5/h5H,2-4H2,1H3',
        'inchi_key': 'LRHPLDYGYMQRHN-UHFFFAOYSA-N',
        'smile': 'CCCCO',
        'canonical_smile': 'CCCCO',
        'cas_number': '71-36-3',
    },
    'butanol': 'butan-1-ol',  # Reference to 'butan-1-ol'
    'butanal': {
        'pub_chem_id': 261,
        'iupac_name': 'butanal',
        'molecular_formula': 'C4H8O',
        'molecular_mass': 72.057514874,
        'molar_mass': 72.11,
        'monoisotopic_mass': 72.057514874,
        'inchi': 'InChI=1S/C4H8O/c1-2-3-4-5/h4H,2-3H2,1H3',
        'inchi_key': 'ZTQSAGDEMFDKMZ-UHFFFAOYSA-N',
        'smile': 'CCCC=O',
        'canonical_smile': 'CCCC=O',
    },
    'butyraldehyde': 'butanal',  # Reference to 'butanal'
    'pent-1-ene': {
        'pub_chem_id': 8004,
        'iupac_name': 'pent-1-ene',
        'molecular_formula': 'C5H10',
        'molecular_mass': 70.078250319,
        'molar_mass': 70.13,
        'monoisotopic_mass': 70.078250319,
        'inchi': 'InChI=1S/C5H10/c1-3-5-4-2/h3H,1,4-5H2,2H3',
        'inchi_key': 'YWAKXRMUMFPDSH-UHFFFAOYSA-N',
        'smile': 'CCCC=C',
        'canonical_smile': 'CCCC=C',
    },
    'n-pentene': 'pent-1-ene',
    'npentene': 'pent-1-ene',
    'ethanol': {
        'pub_chem_id': 702,
        'iupac_name': 'ethanol',
        'molecular_formula': 'C2H6O',
        'molar_mass': 46.07,
        'inchi': 'InChI=1S/C2H6O/c1-2-3/h3H,2H2,1H3',
        'inchi_key': 'LFQSCWFLJHTTHZ-UHFFFAOYSA-N',
        'cas_number': '64-17-5',
    },
    'EtOH': 'ethanol',  # Reference to 'ethanol'
    'etoh': 'ethanol',  # Reference to 'ethanol'
    'CH3CH2OH': 'ethanol',  # Reference to 'ethanol'
    'ch3ch2oh': 'ethanol',  # Reference to 'ethanol'
    'methanol': {
        'pub_chem_id': 887,
        'iupac_name': 'methanol',
        'molecular_formula': 'CH4O',
        'molar_mass': 32.04,
        'inchi': 'InChI=1S/CH4O/c1-2/h2H,1H3',
        'inchi_key': 'OKKJLVBELUTLKV-UHFFFAOYSA-N',
        'cas_number': '67-56-1',
    },
    'methyl alcohol': 'methanol',  # Reference to 'methanol'
    'MeOH': 'methanol',  # Reference to 'methanol'
    'meoh': 'methanol',  # Reference to 'methanol'
    'CH3OH': 'methanol',  # Reference to 'methanol'
    'formic acid': {
        'pub_chem_id': 284,
        'iupac_name': 'formic acid',
        'molecular_formula': 'CH2O2',
        'molar_mass': 46.03,
        'inchi': 'InChI=1S/CH2O2/c2-1-3/h1H,(H,2,3)',
        'inchi_key': 'BDAGIHXWWSANSR-UHFFFAOYSA-N',
        'cas_number': '64-18-6',
    },
    'HCOOH': 'formic acid',  # Reference to 'formic acid'
    'hcooh': 'formic acid',  # Reference to 'formic acid'
    'methanoic acid': 'formic acid',  # Reference to 'formic acid'
    'prop-2-enal': {
        'pub_chem_id': 7847,
        'iupac_name': 'prop-2-enal',
        'molecular_formula': 'C3H4O',
        'molar_mass': 56.06,
        'inchi': 'InChI=1S/C3H4O/c1-2-3-4/h2-3H,1H2',
    },
    'acrolein': 'prop-2-enal',  # Reference to 'prop-2-enal'
    'C3H4O': 'prop-2-enal',  # Reference to 'prop-2-enal'
    'c3h4o': 'prop-2-enal',  # Reference to 'prop-2-enal'
    'propanal': {
        'pub_chem_id': 527,
        'iupac_name': 'propanal',
        'molecular_formula': 'C3H6O',
        'molar_mass': 58.08,
        'inchi': 'InChI=1S/C3H6O/c1-2-3-4/h3H,2H,1H3',
    },
    'propionaldehyde': 'propanal',  # Reference to 'propanal'
    'propan-2-one': {
        'pub_chem_id': 180,
        'iupac_name': 'propan-2-one',
        'molecular_formula': 'C3H6O',
        'molar_mass': 58.08,
        'inchi': 'InChI=1S/C3H6O/c1-3(2)2/h1-2H3',
        'inchi_key': 'CSCPPACGZOOCGX-UHFFFAOYSA-N',
        'cas_number': '67-64-1',
    },
    'acetone': 'propan-2-one',  # Reference to 'propan-2-one'
    'propan-2-ol': {
        'pub_chem_id': 3776,
        'iupac_name': 'propan-2-ol',
        'molecular_formula': 'C3H8O',
        'molar_mass': 60.10,
        'inchi': 'InChI=1S/C3H8O/c1-3(2)4/h3-4H,1-2H3',
        'inchi_key': 'KZBMWSRVAYFASW-UHFFFAOYSA-N',
        'cas_number': '67-63-0',
    },
    '2-propanol': 'propan-2-ol',  # Reference to 'propan-2-ol'
    'isopropanol': 'propan-2-ol',  # Reference to 'propan-2-ol'
    'prop-2-en-1-ol': {
        'pub_chem_id': 7858,
        'iupac_name': 'prop-2-en-1-ol',
        'molecular_formula': 'C3H6O',
        'molar_mass': 58.08,
        'inchi': 'InChI=1S/C3H6O/c1-2-3-4/h2,4H,1,3H2',
        'inchi_key': 'QWVGKYWNOKOFNN-UHFFFAOYSA-N',
        'cas_number': '107-18-6',
    },
    'allyl alcohol': 'prop-2-en-1-ol',  # Reference to 'allyl alcohol'
    'allylalcohol': 'prop-2-en-1-ol',  # Reference to 'prop-2-en-1-ol'
    'propan-1-ol': {
        'pub_chem_id': 1031,
        'iupac_name': 'propan-1-ol',
        'molecular_formula': 'C3H8O',
        'molar_mass': 60.10,
        'inchi': 'InChI=1S/C3H8O/c1-2-3-4/h4H,2-3H2,1H3',
        'inchi_key': 'XNWBBONJTDVZCF-UHFFFAOYSA-N',
        'cas_number': '71-23-8',
    },
    '1-propanol': 'propan-1-ol',  # Reference to 'propan-1-ol'
    'n-propanol': 'propan-1-ol',  # Reference to 'propan-1-ol'
    'npropanol': 'propan-1-ol',  # Reference to 'propan-1-ol'
    'propanol': 'propan-1-ol',  # Reference to 'propan-1-ol'
    'methyl acetate': {
        'pub_chem_id': 6589,
        'iupac_name': 'methyl acetate',
        'molecular_formula': 'C3H6O2',
        'molecular_mass': 74.03677943,
        'molar_mass': 74.08,
        'monoisotopic_mass': 74.03677943,
        'inchi': 'InChI=1S/C3H6O2/c1-3(4)5-2/h1-2H3',
        'inchi_key': 'KXKVLQRXCPHEJC-UHFFFAOYSA-N',
        'smile': 'CC(=O)OC',
        'canonical_smile': 'CC(=O)OC',
    },
    'methylacetate': 'methyl acetate',  # Reference to 'methyl acetate'
    'ethyl acetate': {
        'pub_chem_id': 8857,
        'iupac_name': 'ethyl acetate',
        'molecular_formula': 'C4H8O2',
        'molecular_mass': 88.052429494,
        'molar_mass': 88.11,
        'monoisotopic_mass': 88.052429494,
        'inchi': 'InChI=1S/C4H8O2/c1-3-6-4(2)5/h3H2,1-2H3',
        'inchi_key': 'XEKOWRVHYACXOJ-UHFFFAOYSA-N',
        'smile': 'CCOC(=O)C',
        'canonical_smile': 'CCOC(=O)C',
    },
    'ethylacetate': 'ethyl acetate',  # Reference to 'ethyl acetate'
    '1-butene': {
        'pub_chem_id': 7844,
        'iupac_name': 'but-1-ene',
        'molecular_formula': 'C4H8',
        'molar_mass': 56.11,
        'inchi': 'InChI=1S/C4H8/c1-3-4-2/h3H,1,4H2,2H3',
    },
    'butylene': '1-butene',  # Reference to '1-butene'
    'n-Butene': '1-butene',  # Reference to '1-butene'
    'nbutene': '1-butene',  # Reference to '1-butene'
    'butene': {
        'iupac_name': 'butene',
        'molecular_formula': 'C4H8',
    },  # unspecified isomer
    'nictbutene': 'butene',  # Reference to 'butene', unspecified isomer
    'furan': {
        'pub_chem_id': 8029,
        'iupac_name': 'furan',
        'molecular_formula': 'C4H4O',
        'molar_mass': 68.07,
        'inchi': 'InChI=1S/C4H4O/c1-2-4-5-3-1/h1-4H',
        'inchi_key': 'NNTHXFGICWXRPZ-UHFFFAOYSA-N',
        'cas_number': '110-00-9',
    },
    'furfuran': 'furan',  # Reference to 'furan'
    'furan-2,5-dione': {
        'pub_chem_id': 7923,
        'iupac_name': 'furan-2,5-dione',
        'molecular_formula': 'C4H2O3',
        'molar_mass': 98.06,
        'inchi': 'InChI=1S/C4H2O3/c5-3-1-2-4(6)7-3/h1-2H',
    },
    '2,5-dihydrofuran': {
        'pub_chem_id': 15570,
        'iupac_name': '2,5-dihydrofuran',
        'molecular_formula': 'C4H6O',
        'molecular_mass': 70.041864811,
        'molar_mass': 70.09,
        'monoisotopic_mass': 70.041864811,
        'inchi': 'InChI=1S/C4H6O/c1-2-4-5-3-1/h1-2H,3-4H2',
        'inchi_key': 'ARGCQEVBJHPOGB-UHFFFAOYSA-N',
        'smile': 'C1C=CCO1',
        'canonical_smile': 'C1C=CCO1',
    },
    '25-dihydrofuran': '2,5-dihydrofuran',  # Reference to '2,5-dihydrofuran'
    'maleic anhydride': 'furan-2,5-dione',  # Reference to 'furan-2,5-dione'
    'MAN': 'furan-2,5-dione',  # Reference to 'furan-2,5-dione'
    'man': 'furan-2,5-dione',  # Reference to 'furan-2,5-dione'
    'p>=5c': {
        'name': 'p>=5c',
        'iupac_name': 'C5+',
    },  # unspecified longer hydrocarbon
    'C2-C4': {
        'name': 'C2-C4',
        'iupac_name': 'C2-C4',
    },  # unspecified C2-C4 hydrocarbon
    'c2-c4': 'C2-C4',  # Reference to 'C2-C4'
}



from nomad.config.models.ui import (
    App,
    Axis,
    Column,
    Dashboard,
    Layout,
    Menu,
    MenuItemCustomQuantities,
    MenuItemHistogram,
    MenuItemPeriodicTable,
    MenuItemTerms,
    SearchQuantities,
    WidgetPeriodicTable,
    WidgetScatterPlot,
    WidgetTerms,
)

heterogeneous_catalysis_app = App(
    label='Heterogeneous Catalysis',
    path='heterogeneouscatalyst',
    category='Use Cases',
    description='Search heterogeneous catalysts',
    readme="""This page allows you to search **catalyst and catalysis data**
        within NOMAD. The filter menu on the left and the shown
        default columns are specifically designed for Heterogeneous Catalyst
        exploration. The dashboard directly shows useful
        interactive statistics about the data.\n

        In order to generate a custom scatterplot, click on the "+ Scatterplot" button.
        You can then select the x and y quantities by starting to type the property name
        and selecting the appropriate line from the dropdown menus. If the property 
        belongs to
        a repeated quantity, such as e.g. the reactants in a reaction, you can select 
        the either select a specific index, or use the wildcard "*" to select all 
        indices, or
        indicate the specific name of the reactant. If you e.g. want to generate an 
        S-X plot
        for the conversion of the reactant A vs. selectivity to the product B, you can
        select the x-quantity as 
        `results.properties.catalytic.reaction.reactants[? name=="A"].conversion`
        and the y-quantity as 
        `results.properties.catalytic.reaction.products[? name=="B"].selectivity`.
        Be aware that the IUPAC names have to be used for the reactants and products 
        (except for water and ammonia).""",
    filters_locked={'quantities': ['results.properties.catalytic']},
    # search_syntaxes= {
    #     "exclude": ["free_text"]
    # },
    columns=[
        Column(quantity='entry_name', selected=True),
        Column(
            quantity='results.properties.catalytic.reaction.name',
            label='Reaction name',
            selected=True,
        ),
        Column(
            quantity='results.properties.catalytic.catalyst.catalyst_type',
            selected=True,
        ),
        Column(
            quantity='results.properties.catalytic.catalyst.preparation_method',
            label='Preparation',
            selected=True,
        ),
        Column(
            quantity='results.properties.catalytic.catalyst.surface_area',
            format={'decimals': 2, 'mode': 'standard'},
            unit='m^2/g',
            label='Surface area',
        ),
        Column(quantity='results.material.elements'),
        Column(quantity='results.properties.catalytic.catalyst.catalyst_name'),
        Column(
            quantity='results.properties.catalytic.reaction.reactants.name',
            label='Reactants',
        ),
        Column(
            quantity='results.properties.catalytic.reaction.products.name',
            label='Products',
        ),
        Column(
            quantity='results.properties.catalytic.reaction.type',
            label='Reaction type',
        ),
        Column(quantity='references'),
        Column(quantity='results.material.chemical_formula_hill', label='Formula'),
        Column(quantity='results.material.structural_type'),
        Column(quantity='results.eln.lab_ids'),
        Column(quantity='results.eln.sections'),
        Column(quantity='results.eln.methods'),
        Column(quantity='results.eln.tags'),
        Column(quantity='results.eln.instruments'),
        Column(quantity='entry_type'),
        Column(quantity='mainfile'),
        Column(quantity='upload_create_time', label='Upload time'),
        Column(quantity='authors'),
        Column(quantity='comment'),
        Column(quantity='datasets'),
        Column(quantity='published', label='Access'),
        Column(
            quantity='data.datetime#nomad_catalysis.schema_packages.catalysis.CatalystSample',
            label='Sample preparation time',
        ),
        Column(
            quantity='data.datetime#nomad_catalysis.schema_packages.catalysis.CatalyticReaction',
            label='Measurement time',
        ),
    ],
    search_quantities=SearchQuantities(
        include=['*#nomad_catalysis.schema_packages.catalysis.Cat*'],
    ),
    menu=Menu(
        title='Heterogeneous Catalysis',
        items=[
            Menu(
                title='Catalyst Materials',
                indentation=1,
            ),
            Menu(
                title='Elements / Formula',
                indentation=2,
                size='xxl',
                items=[
                    MenuItemPeriodicTable(search_quantity='results.material.elements'),
                    MenuItemTerms(
                        search_quantity='results.material.chemical_formula_hill',
                        width=6,
                        options=0,
                    ),
                    MenuItemTerms(
                        search_quantity='results.material.chemical_formula_iupac',
                        width=6,
                        options=0,
                    ),
                    MenuItemTerms(
                        search_quantity='results.material.chemical_formula_reduced',
                        width=6,
                        options=0,
                    ),
                    MenuItemTerms(
                        search_quantity='results.material.chemical_formula_anonymous',
                        width=6,
                        options=0,
                    ),
                    MenuItemHistogram(
                        x={'search_quantity': 'results.material.n_elements'}
                    ),
                ],
            ),
            Menu(
                title='Catalyst Properties',
                indentation=2,
                size='md',
                items=[
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.catalyst.catalyst_type'
                    ),
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.catalyst.support'
                    ),
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.catalyst.preparation_method'
                    ),
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.catalyst.catalyst_name'
                    ),
                    MenuItemTerms(
                        search_quantity='data.form#nomad_catalysis.schema_packages.catalysis.CatalystSample'
                    ),
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.catalyst.characterization_methods'
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.catalyst.surface_area', # noqa: E501
                            'unit': 'm^2/g',
                        },
                        autorange=False,
                    ),
                ],
            ),
            Menu(
                title='Reactions',
                size='md',
                indentation=1,
                items=[
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.reaction.type'
                    ),
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.reaction.name'
                    ),
                ],
            ),
            Menu(
                title='Reactants',
                indentation=2,
                size='md',
                items=[
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.reaction.reactants.name'
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reactants.conversion' # noqa: E501
                        }
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reactants.mole_fraction_in' # noqa: E501
                        }
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reactants.mole_fraction_out' # noqa: E501
                        }
                    ),
                ],
            ),
            Menu(
                title='Products',
                indentation=2,
                size='md',
                items=[
                    MenuItemTerms(
                        search_quantity='results.properties.catalytic.reaction.products.name'
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.products.selectivity' # noqa: E501
                        }
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.products.mole_fraction_out' # noqa: E501
                        }
                    ),
                ],
            ),
            Menu(
                title='Reaction Conditions',
                indentation=2,
                size='md',
                items=[
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reaction_conditions.temperature' # noqa: E501
                        }
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reaction_conditions.pressure', # noqa: E501
                            'unit': 'bar',
                        }
                    ),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'results.properties.catalytic.reaction.reaction_conditions.weight_hourly_space_velocity',  # noqa: E501
                            'unit': 'ml/(g*hr)',
                        }
                    ),
                ],
            ),
            Menu(
                title='Author / Dataset',
                size='md',
                items=[
                    MenuItemTerms(search_quantity='authors.name'),
                    MenuItemHistogram(x={'search_quantity': 'upload_create_time'}),
                    MenuItemTerms(search_quantity='datasets.dataset_name'),
                ],
            ),
            Menu(
                title='Electronic Lab Notebook',
                size='md',
                items=[
                    MenuItemTerms(search_quantity='results.eln.sections'),
                    MenuItemTerms(search_quantity='results.eln.methods'),
                    MenuItemHistogram(
                        x={
                            'search_quantity': 'data.datetime#nomad_catalysis.schema_packages.catalysis.CatalyticReaction' # noqa: E501
                        },
                        title='Date of the catalytic reaction test',
                    ),
                    MenuItemTerms(search_quantity='results.eln.tags'),
                    MenuItemTerms(search_quantity='results.eln.instruments'),
                    MenuItemTerms(search_quantity='results.eln.lab_ids'),
                ],
            ),
            Menu(
                title='User Defined Quantities',
                size='xl',
                items=[
                    MenuItemCustomQuantities(),
                ],
            ),
        ],
    ),
    dashboard=Dashboard(
        widgets=[
            WidgetPeriodicTable(
                layout={
                    'lg': Layout(h=8, minH=8, minW=12, w=12, x=0, y=0),
                    'md': Layout(h=8, minH=6, minW=8, w=10, x=0, y=0),
                    'sm': Layout(h=8, minH=8, minW=12, w=12, x=0, y=0),
                    'xl': {'h': 10, 'minH': 8, 'minW': 12, 'w': 12, 'x': 0, 'y': 0},
                    'xxl': {'h': 10, 'minH': 8, 'minW': 12, 'w': 16, 'x': 0, 'y': 0},
                },
                search_quantity='results.material.elements',
                scale='linear',
                title='Elements of the catalyst material',
            ),
            WidgetTerms(
                title='Reaction name',
                layout={
                    'lg': Layout(h=8, minH=3, minW=3, w=6, x=12, y=0),
                    'md': Layout(h=8, minH=3, minW=3, w=4, x=10, y=0),
                    'sm': Layout(h=4, minH=3, minW=3, w=4, x=0, y=8),
                    'xl': Layout(h=10, minH=3, minW=3, w=6, x=12, y=0),
                    'xxl': Layout(h=10, minH=3, minW=3, w=6, x=16, y=0),
                },
                search_quantity='results.properties.catalytic.reaction.name',
            ),
            WidgetTerms(
                title='Reactants',
                layout={
                    'lg': Layout(h=4, minH=3, minW=3, w=6, x=18, y=0),
                    'md': Layout(h=4, minH=3, minW=3, w=4, x=14, y=0),
                    'sm': Layout(h=4, minH=4, minW=3, w=4, x=4, y=8),
                    'xl': Layout(h=5, minH=3, minW=3, w=6, x=18, y=0),
                    'xxl': Layout(h=10, minH=3, minW=3, w=6, x=22, y=0),
                },
                search_quantity='results.properties.catalytic.reaction.reactants.name',
                showinput=True,
                scale='linear',
            ),
            WidgetTerms(
                title='Products',
                layout={
                    'lg': Layout(h=4, minH=3, minW=3, w=6, x=18, y=4),
                    'md': Layout(h=4, minH=3, minW=3, w=4, x=14, y=4),
                    'sm': Layout(h=4, minH=4, minW=3, w=4, x=8, y=8),
                    'xl': Layout(h=5, minH=3, minW=3, w=6, x=18, y=5),
                    'xxl': Layout(h=10, minH=3, minW=3, w=6, x=28, y=0),
                },
                search_quantity='results.properties.catalytic.reaction.products.name',
                showinput=True,
                scale='linear',
            ),
            WidgetScatterPlot(
                title='Reactant concentrations vs. Temperature',
                autorange=True,
                layout={
                    'lg': Layout(h=10, minH=3, minW=8, w=12, x=0, y=8),
                    'md': Layout(h=6, minH=3, minW=8, w=9, x=0, y=8),
                    'sm': Layout(h=6, minH=3, minW=6, w=6, x=0, y=12),
                    'xl': Layout(h=8, minH=3, minW=8, w=12, x=0, y=10),
                    'xxl': Layout(h=8, minH=6, minW=8, w=12, x=0, y=10),
                },
                x=Axis(
                    search_quantity='results.properties.catalytic.reaction.reactants[*].mole_fraction_in',
                    title='Molar fraction of reactants',
                ),
                y=Axis(
                    search_quantity='results.properties.catalytic.reaction.reaction_conditions.temperature'
                ),
                color='results.properties.catalytic.reaction.reactants[*].name',
                size=1000,
            ),
            WidgetScatterPlot(
                title='Temperature vs. Conversion',
                autorange=True,
                layout={
                    'lg': Layout(h=10, minH=3, minW=3, w=12, x=12, y=8),
                    'md': Layout(h=6, minH=3, minW=3, w=9, x=9, y=8),
                    'sm': Layout(h=6, minH=3, minW=3, w=6, x=6, y=12),
                    'xl': Layout(h=8, minH=3, minW=3, w=12, x=12, y=10),
                    'xxl': Layout(h=8, minH=3, minW=3, w=12, x=12, y=10),
                },
                x=Axis(
                    search_quantity='results.properties.catalytic.reaction.reaction_conditions.temperature'
                ),
                y=Axis(
                    search_quantity='map(&conversion, results.properties.catalytic.reaction.reactants[*])', # noqa: E501
                    title='Conversion (%)',
                ),
                color='map(&name, results.properties.catalytic.reaction.reactants[*])',
                size=1000,
            ),
        ]
    ),
)




import numpy as np
import pandas as pd
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.basesections import (
    CompositeSystemReference,
    ElementalComposition,
    SectionReference,
)
from nomad.parsing import MatchingParser
from nomad.units import ureg

from nomad_catalysis.parsers.utils import create_archive
from nomad_catalysis.schema_packages.catalysis import (
    CatalysisCollectionParserEntry,
    CatalystSample,
    CatalyticReaction,
    CatalyticReactionData,
    Preparation,
    ProductData,
    RatesData,
    RawFileData,
    ReactantData,
    ReactionConditionsData,
    ReactorFilling,
    ReactorSetup,
    Reagent,
    SurfaceArea,
)


def get_time_unit(string) -> any:
    """
    This function extracts the time unit (h/min/s) from a string.
    It returns a ureg.Quantity object with the time unit.
    """
    if 'h' in string:
        return ureg.hour
    elif 's' in string:
        return ureg.second
    elif 'min' in string:
        return ureg.minute
    else:
        raise ValueError('Time unit not recognized.')

def get_mass_unit(string) -> any:
    """
    This function extracts the mass unit (g/mg/kg) from a string.
    It returns a ureg.Quantity object with the mass unit.
    """
    string = string.strip('([])').casefold()
    if 'mg' in string:
        return ureg.milligram
    elif 'kg' in string:
        return ureg.kilogram
    elif string in ['g', 'gram']:
        return ureg.gram
    else:
        raise ValueError('Mass unit not recognized.')
    

class CatalysisParser(MatchingParser):
    def parse(
        self,
        mainfile: str,
        archive: EntryArchive,
        logger=None,
        child_archives: dict[str, EntryArchive] = None,
    ) -> None:
        filename = mainfile.split('/')[-1]
        name = filename.split('.')[0]
        logger.info(f' Catalysis Parser called {filename}')

        catalytic_reaction = CatalyticReaction(
            data_file=filename,
        )

        archive.data = RawFileData(
            measurement=create_archive(
                catalytic_reaction, archive, f'{name}.archive.json'
            )
        )
        archive.metadata.entry_name = f'{name} data file'


class CatalysisCollectionParser(MatchingParser):
    def unify_columnnames(self, data_frame) -> pd.DataFrame:
        """
        This function unifies the column names of the data frame to a common format
        by renaming the columns to a standard format that is used in the rest of
        the code *for a sample entry*. Here, the column 'name' is droped if a 
        column 'catalyst' is present, as the focus is on the sample entry which is
        the catalyst.
        """

        for col in data_frame.columns:
            if col in ['catalyst name','catalyst_name', 'catalyst']:
                try:
                    data_frame.drop(columns=['name'], inplace=True)
                except KeyError:
                    pass
                data_frame.rename(columns={col: 'name'}, inplace=True)
            if col in ['storing institution', 'storing_institute']:
                data_frame.rename(columns={col: 'storing_institution'}, inplace=True)
            if col in ['date', 'sample date']:
                data_frame.rename(columns={col: 'datetime'}, inplace=True)
            if col in ['lab-id','sample_id', 'catalyst_id']:
                data_frame.rename(columns={col: 'lab_id'}, inplace=True)
            if col in ['surface_area_method']:
                data_frame.rename(
                    columns={col: 'method_surface_area_determination'}, inplace=True
                )
            if col in ['surface_area (m2/g)']:
                data_frame.rename(
                    columns={col: 'surface_area'}, inplace=True
                )
            if col in ['preparation']:
                data_frame.rename(columns={col: 'preparation_method'}, inplace=True)
            if col in ['comment', 'comments']:
                data_frame.rename(columns={col: 'description'}, inplace=True)
                
        return data_frame
    
    def check_zero_elements(
        self, el, logger) -> bool:
        """
        Checks if the element has a zero atomic or mass fraction.
        If it does, this will not be written in the results section of the reaction.
        """
        if el.atomic_fraction == 0.0 or el.mass_fraction == 0.0:
            logger.info(
                f'''{el.element} has a zero atomic or mass fraction and will not be
                appended to the elemental composition.'''
            )
            return True
        else:
            return False
    
    def extract_elemental_composition(self, row, catalyst_sample, logger) -> None:
        """
        This function extracts the elemental composition from a row of the data frame.
        It returns an ElementalComposition object with the element and its mass and atom
        fractions.
        """
        if 'elements' in row.keys():
            elements = row['elements'].split(',')
        elif 'element' in row.keys():
            elements = []
            elements.append(row['element'])
            if 'element.1' in row.keys():
                for key in row.keys():
                    if key.startswith('element.'):
                        elements.append(row[key])

        for m, element in enumerate(elements):
            elemental_composition = ElementalComposition(
                element=element.strip(),
            )
            try:
                mass_fractions = row['mass_fractions'].split(',')
                elemental_composition.mass_fraction = float(
                    mass_fractions[m]
                )
            except KeyError:
                pass
            try:
                atom_fractions = row['atom_fractions'].split(',')
                elemental_composition.atom_fraction = float(
                    atom_fractions[m]
                )
            except KeyError:
                pass
            zero_element = self.check_zero_elements(
                elemental_composition, logger
            )
            if zero_element:
                continue
            catalyst_sample.elemental_composition.append(
                elemental_composition
            )

    def extract_reaction_feed(self, row, logger) -> ReactionConditionsData:
        """
        This function extracts the reaction feed from a row of the data frame.
        It returns a ReactionConditionsData object with the feed information.
        """
        feed = ReactionConditionsData()
        for key in row.keys():
            col_split = key.split(' ')

            if col_split[0].casefold() == 'set_temperature':
                if 'k' in col_split[1]:
                    feed.set_temperature = [np.nan_to_num(row[key])]
                elif 'c' in col_split[1]:
                    feed.set_temperature = [float(np.nan_to_num(row[key]))+273.15]

            if col_split[0].casefold() == 'tos' or (
                col_split[0].casefold() == 'time'):
                unit = get_time_unit(col_split[1])
                feed.time_on_stream = [np.nan_to_num(row[key])] * unit

            if col_split[0] == 'ghsv':
                if '1/h' in col_split[1] or 'h^-1' in col_split[1]:
                    feed.gas_hourly_space_velocity = (
                        [np.nan_to_num(row[key])] * ureg.hour**-1
                    )
                else:
                    logger.warning('Gas hourly space velocity unit not recognized.')

            if col_split[0] == 'whsv':
                if 'ml/g/h' in col_split[1] or 'ml/(g*h)' in col_split[1]:
                    feed.weight_hourly_space_velocity = (
                        [np.nan_to_num(row[key])]
                        * ureg.milliliter
                        / (ureg.gram * ureg.hour)
                    )

            if ((col_split[0] == 'vflow' or col_split[0] == 'flow_rate')
                and ('ml/min' in col_split[1] or 'mln' in col_split[1])):
                    feed.set_total_flow_rate = (
                        [np.nan_to_num(row[key])] * ureg.milliliter / ureg.minute
                    )

            if col_split[0] == 'set_pressure' and 'bar' in col_split[1]:
                feed.set_pressure = [np.nan_to_num(row[key])] * ureg.bar

        return feed
    
    def extract_catalytic_results(self, row, logger) -> CatalyticReactionData:
        """
        This function extracts the catalytic results from a row of the data frame.
        It returns a CatalyticReactionData object with the results information.
        """
        cat_data = CatalyticReactionData()
        
        for key in row.keys():
            col_split = key.split(' ')

            if key == 'c-balance':
                cat_data.c_balance = [np.nan_to_num(row[key])]
            elif col_split[0] == 'c-balance' and ('%' in col_split[1]):
                cat_data.c_balance = [np.nan_to_num(row[key])] / 100

            if col_split[0] == 'temperature':
                if 'k' in col_split[1]:
                    cat_data.temperature = [np.nan_to_num(row[key])]
                elif 'c' in col_split[1]:
                    cat_data.temperature = [float(np.nan_to_num(row[key]))+273.15]
                else:
                    logger.warning('Temperature unit not recognized.')

            if col_split[0] == 'pressure':
                if 'bar' in col_split[1]:
                    cat_data.pressure = [np.nan_to_num(row[key])] * ureg.bar
                else:
                    logger.warning('Pressure unit not recognized.')

            if col_split[0] == 'tos' or col_split[0] == 'time':
                unit = get_time_unit(col_split[1])
                cat_data.time_on_stream = [np.nan_to_num(row[key])] * unit

        return cat_data
    
    def extract_reactor_setup(self, row, logger) -> ReactorSetup:
        """
        This function extracts the reactor setup information from a row of the data
        frame. It returns a ReactorSetup object with the reactor setup information.
        """
        reactor_setup = ReactorSetup()
        
        for key in row.keys():
            col_split = key.split(' ')

            if key == 'reactor_type':
                    reactor_setup.reactor_type = row[key]
            if key.startswith('reactor_volume'):
                unit = col_split[1].strip('()')
                try:
                    reactor_setup.reactor_volume = ureg.Quantity(
                        np.nan_to_num(row[key]), unit
                    )
                except Exception as e:
                    logger.warning(f"""Reactor volume unit {unit} not recognized. 
                                Error: {e}""")
            if key.startswith('reactor_diameter'):
                unit = col_split[1].strip('()')
                try:
                    reactor_setup.reactor_diameter = ureg.Quantity(
                        np.nan_to_num(row[key]), unit
                    )
                except Exception as e:
                    logger.warning(f"""Reactor diameter unit {unit} not recognized. 
                                Error: {e}""")
            if key == 'reactor_lab_id':
                reactor_setup.lab_id = row[key]
            if key == 'reactor_name':
                reactor_setup.name = row[key]

        return reactor_setup

    def extract_pretreatment(self, row, logger) -> ReactionConditionsData:  # noqa: PLR0912, PLR0915
        """
        This function extracts the pretreatment information from a row of the data
        frame. It returns a ReactorFilling object with the pretreatment information.
        """
        pretreatment = ReactionConditionsData()
        pretreatment_reagents = []
        logger.info('Extracting pretreatment information from the data frame')
        for key in row.keys():
            col_split = key.split(' ')

            if col_split[0] == 'pretreatment':
                if col_split[1].startswith('set_temperature'):
                    if pretreatment.set_temperature is None:
                        pretreatment_temperature = [row[key]]
                    else:
                        pretreatment_temperature.append(row[key])
                    if 'c' in col_split[2]:
                        pretreatment_temperature_np = np.array(pretreatment_temperature)
                        pretreatment.set_temperature = (
                        pretreatment_temperature_np + 273.15)
                    elif 'k' in col_split[2]:
                        pretreatment.set_temperature= pretreatment_temperature
                    else:
                        logger.warning('Temperature unit not recognized.')

                if col_split[1].startswith('time'):
                    if pretreatment.time_on_stream is None:
                        tos = []
                    if len(col_split) == 3:  # noqa: PLR2004
                        unit = get_time_unit(col_split[2])
                    else:
                        logger.error('Time unit missing.')
                    tos.append(np.nan_to_num(row[key]))
                    pretreatment.time_on_stream = tos * unit
                if col_split[1].startswith('set_pressure'):
                    if not pretreatment.set_pressure:
                        pretreatment.set_pressure = []
                    if 'bar' in col_split[2]:
                        pretreatment.set_pressure.append(
                            np.nan_to_num(row[key])) # * ureg.bar
                    else:
                        logger.warning('Pressure unit not recognized.')
                if col_split[1].startswith('set_flow_rate'):
                    if not pretreatment.set_total_flow_rate:
                        pretreatment.set_total_flow_rate = []
                    if 'ml/min' in col_split[2] or 'mln' in col_split[2]:
                        total_flow=np.append(pretreatment.set_total_flow_rate.to('milliliter/minute').magnitude,
                                    row[key])
                        pretreatment.set_total_flow_rate = (
                            total_flow * ureg.milliliter / ureg.minute
                        )
                    else:
                        logger.warning(f'Flow rate unit not recognized from {key}.')
                if col_split[1].startswith('gas_flow'):
                    try:
                        if len(col_split) == 4 and (  # noqa: PLR2004
                            'ml/min' in col_split[3] or 'mln' in col_split[3]):
                            if col_split[2] not in pretreatment_reagents :
                                reagent = Reagent(
                                    name=col_split[2],
                                    flow_rate=(
                                            [np.nan_to_num(row[key])]
                                            * ureg.milliliter
                                            / ureg.minute
                                        ),
                                )
                                pretreatment_reagents.append(col_split[2])
                                pretreatment.reagents.append(reagent)
                            else:
                                index = pretreatment_reagents.index(col_split[2])
                                gas_flow = np.append(
                                    pretreatment.reagents[index].flow_rate.to('milliliter/minute').magnitude,
                                    row[key]
                                )
                                pretreatment.reagents[index].flow_rate = (
                                    gas_flow * ureg.milliliter / ureg.minute
                                )
                        else:
                            logger.warning(f'unit in {key} missing or not recognized.')
                            
                    except KeyError:
                        logger.warning(f'Gas flow for {key} not recognized.')

        return pretreatment

    def extract_reaction_entries(self, data_frame, archive, logger) -> None:  # noqa: PLR0912, PLR0915
        "This function extracts information for catalytic reaction entries with a"
        "single measurement from the data frame and adds them to the archive."
        reactions = []
        
        data_frame.dropna(axis=1, how='all', inplace=True)
        for n, row in data_frame.iterrows():
            row.dropna(inplace=True)

            reaction = CatalyticReaction()
            reactor_filling = ReactorFilling()
            sample = CompositeSystemReference()

            reagents = []
            reagent_names = []
            products = []
            product_names = []
            conversions = []
            conversion_names = []
            rates = []

            if 'reaction_type' in row.keys():
                reaction.reaction_type = []
                types= row['reaction_type'].split(',')
                if isinstance(types, list):
                    reaction.reaction_type.extend(types)
                else:
                    reaction.reaction_type.append(types)
            for key in [
                    'datetime',
                    'lab_id',
                    'description',
                    'reaction_name',
                    'experimenter',
                    'location',
                ]:
                if key in row.keys():
                    setattr(reaction, key, row[key])

            if 'datafile' in row.keys():
                reaction.data_file = row['datafile']

                reactions.append(
                        create_archive(
                            reaction,
                            archive,
                            f'{row["name"]}_catalytic_reaction.archive.json',
                        )
                    )
                continue

            feed = self.extract_reaction_feed(row, logger)
            cat_data = self.extract_catalytic_results(row, logger)
            reactor_setup = self.extract_reactor_setup(row, logger)
            pretreatment = self.extract_pretreatment(row, logger)

            for key in row.keys():
                
                col_split = key.split(' ')

                if key in ['catalyst', 'catalyst_name']:
                    setattr(sample, 'name', row[key])
                    setattr(reactor_filling, 'catalyst_name', str(row[key]))
                if key in ['sample_id', 'catalyst_id']:
                    setattr(sample, 'lab_id', row[key])
                
                # if len(col_split) < 2:  # noqa: PLR2004
                #     continue

                if col_split[0].casefold() == 'x':
                    if len(col_split) == 3 and ('%' in col_split[2]):  # noqa: PLR2004
                        try:
                            gas_in = [np.nan_to_num(float(row[key])) / 100.]
                        except (ValueError, TypeError) as e:
                            logger.warning(f"""Non-numeric value for {key}: {row[key]}. 
                                           Error: {e}.""")
                            gas_in = [np.nan]
                    else:
                        gas_in = [row[key]]
                    reagent = Reagent(name=col_split[1], fraction_in=gas_in)
                    reagent_names.append(col_split[1])
                    reagents.append(reagent)

                if col_split[0].casefold() == 'mass':
                    unit=get_mass_unit(col_split[1])
                    try:
                        reactor_filling.catalyst_mass = (row[key] * unit)
                    except Exception as e:
                        logger.warning(f"""Catalyst mass unit {col_split[1]} not 
                                       recognized. Error: {e}""")
                if key == 'diluent':
                    reactor_filling.diluent = row[key]
                    for key2 in row.keys():
                        if key2.startswith('diluent_mass'):
                            col_split = key2.split(' ')
                            unit = get_mass_unit(col_split[1])
                            try:
                                reactor_filling.diluent_mass = row[key2] * unit
                            except Exception as e:
                                logger.warning(f"""Diluents mass unit {col_split[1]} not
                                               recognized. Error: {e}""")

                if len(col_split) < 3:  # noqa: PLR2004
                    continue

                if col_split[0] == 'r':  # reaction rate
                    unit = col_split[2].strip('()')
                    unit_conversion = {
                        'mmol/g/h': 'mmol / (g * hour)',
                        'mmol/g/min': 'mmol / (g * minute)',
                        'µmol/g/min': 'µmol / (g * minute)',
                        'mmolg^-1h^-1': 'mmol / (g * hour)',
                    }
                    try:
                        rate = RatesData(
                            name=col_split[1],
                            reaction_rate=ureg.Quantity(
                                [np.nan_to_num(row[key])], 
                                unit_conversion.get(unit, unit)
                            ),
                        )
                    except Exception as e:
                        logger.warning(f"""Reaction rate unit {unit} not recognized. 
                                    Error: {e}""")
                    rates.append(rate)

                if col_split[0] == 'r_specific_mass':  # specific reaction rate
                    unit = col_split[2].strip('()')
                    unit_conversion = {
                        'mol/(h*gmetal': 'mol / (hour * g)',
                    }
                    try:
                        rate = RatesData(
                            name=col_split[1],
                            specific_mass_rate=ureg.Quantity(
                                [np.nan_to_num(row[key])], 
                                unit_conversion.get(unit, unit)
                            ),
                        )
                    except Exception as e:
                        logger.warning(f"""Specific reaction rate per mass unit {unit}
                            not recognized. Error: {e}""")
                    rates.append(rate)

                if col_split[2] != '(%)':
                    continue

                if col_split[0] == 'x_p':  # conversion, based on product detection
                    conversion = ReactantData(
                        name=col_split[1],
                        conversion=[np.nan_to_num(row[key])],
                        conversion_type='product-based conversion',
                        conversion_product_based=[np.nan_to_num(row[key])],
                    )
                    for i, p in enumerate(conversions):
                        if p.name == col_split[1]:
                            conversion = conversions.pop(i)

                    conversion.conversion_product_based = [np.nan_to_num(row[key])]
                    conversion.conversion = [np.nan_to_num(row[key])]
                    conversion.conversion_type = 'product-based conversion'

                    conversion_names.append(col_split[1])
                    conversions.append(conversion)

                if col_split[0] == 'x_r':  # conversion, based on reactant detection
                    try:
                        conversion = ReactantData(
                            name=col_split[1],
                            conversion=[np.nan_to_num(row[key])],
                            conversion_type='reactant-based conversion',
                            conversion_reactant_based=[np.nan_to_num(row[key])],
                            fraction_in=[
                                np.nan_to_num(float(row['x ' + col_split[1] + ' (%)']))
                                / 100],
                        )
                    except KeyError:
                        conversion = ReactantData(
                            name=col_split[1],
                            conversion=[np.nan_to_num(row[key])],
                            conversion_type='reactant-based conversion',
                            conversion_reactant_based=[np.nan_to_num(row[key])],
                            fraction_in=[np.nan_to_num(row['x ' + col_split[1]])],
                        )

                    for i, p in enumerate(conversions):
                        if p.name == col_split[1]:
                            conversion = conversions.pop(i)
                            conversion.conversion_reactant_based = [
                                np.nan_to_num(row[key])]
                    conversions.append(conversion)

                if col_split[0].casefold() == 'x_out':  # concentration out
                    if col_split[1] in reagent_names:
                        if "%" in key:
                            fraction_in = [
                                np.nan_to_num(float(row['x ' + col_split[1] + ' (%)']))
                                / 100
                            ]
                            fraction_out = [np.nan_to_num(row[key]) / 100]
                        else:
                            fraction_in = [np.nan_to_num(row['x ' + col_split[1]])]
                            fraction_out = [np.nan_to_num(row[key])]
                        
                        conversion = ReactantData(
                            name=col_split[1],
                            fraction_in=fraction_in,
                            fraction_out=fraction_out,
                        )
                        
                        conversions.append(conversion)
                    else:
                        product = ProductData(
                            name=col_split[1],
                            fraction_out=[np.nan_to_num(row[key])] / 100,
                        )
                        products.append(product)
                        product_names.append(col_split[1])

                if col_split[0].casefold() == 's_p':  # selectivity
                    product = ProductData(
                        name=col_split[1], selectivity=[np.nan_to_num(row[key])]
                    )
                    for i, p in enumerate(products):
                        if p.name == col_split[1]:
                            product = products.pop(i)
                            product.selectivity = [np.nan_to_num(row[key])]
                            break
                    products.append(product)
                    product_names.append(col_split[1])

                if col_split[0].casefold() == 'y':  # product yield
                    product = ProductData(
                        name=col_split[1], product_yield=[np.nan_to_num(row[key])]
                    )
                    for i, p in enumerate(products):
                        if p.name == col_split[1]:
                            product = products.pop(i)
                            product.product_yield = [np.nan_to_num(row[key])]
                            break
                    products.append(product)
                    product_names.append(col_split[1])
                
            reaction.samples = []
            reaction.samples.append(sample)

            cat_data.products = products
            if conversions != []:
                cat_data.reactants_conversions = conversions
            if rates != []:
                cat_data.rates = rates

            feed.reagents = reagents

            reaction.reaction_conditions = feed
            reaction.results = []
            reaction.results.append(cat_data)

            if reactor_filling != []:
                reaction.reactor_filling = reactor_filling
            if reactor_setup:
                reaction.instruments = []
                reaction.instruments.append(reactor_setup)
            if pretreatment:
                reaction.pretreatment = pretreatment

            reactions.append(
                create_archive(
                    reaction,
                    archive,
                    f'{row["name"]}_catalytic_reaction.archive.json',
                )
            )
        reaction_references = []
        for n, reaction in enumerate(reactions):
            reaction_ref = SectionReference(
                reference=reaction,
                name=data_frame["name"][n]
            )
            reaction_references.append(reaction_ref)
        archive.data.measurements = reaction_references

    def extract_sample_entries(self, data_frame, archive, logger
                               ) -> list[CatalystSample]:
        """ This function extracts information for catalyst sample entries from the
        data frame and adds them to the archive. It returns a list of CatalystSample
        objects."""
        logger.info('Extracting sample entries from the data frame')

        samples = []
        for n, row in data_frame.iterrows():
            row.dropna(inplace=True)
            catalyst_sample = CatalystSample()
            surface = SurfaceArea()
            preparation_details = Preparation()
            
            for key in [
                    'name',
                    'storing_institution',
                    'datetime',
                    'lab_id',
                    'form',
                    'support',
                    'description',
                    'formula_descriptive',
                ]:
                if key in row.keys():
                    setattr(catalyst_sample, key, row[key])
            if 'catalyst_type' in row.keys():
                    catalyst_sample.catalyst_type = []
                    catalyst_sample.catalyst_type.extend([row['catalyst_type']])
                    #setattr(catalyst_sample, key, row['catalyst_type'])
            if 'elements' in row.keys() or 'element' in row.keys():
                self.extract_elemental_composition(row, catalyst_sample, logger)
                    
            for key in ['preparation_method', 'preparator', 'preparing_institution']:
                if key in row.keys():
                    setattr(preparation_details, key, row[key])
            for key in [
                    'surface_area',
                    'method_surface_area_determination',
                    'dispersion',
                ]:
                if key in row.keys():
                    setattr(surface, key, row[key])

            if preparation_details.m_to_dict():
                catalyst_sample.preparation_details = preparation_details
            if surface.m_to_dict():
                catalyst_sample.surface = surface

            samples.append(
                create_archive(
                    catalyst_sample,
                    archive,
                    f'{row["name"]}_catalyst_sample.archive.json',
                )
            )
        return samples

    
    def parse(
        self,
        mainfile: str,
        archive: EntryArchive,
        logger=None,
        child_archives: dict[str, EntryArchive] = None,
    ) -> None:
        
        logger.info('Catalysis Collection Parser called')

        filename = mainfile.split('/')[-1]
        name = filename.split('.')

        archive.data = CatalysisCollectionParserEntry(
            data_file=filename,
        )
        archive.metadata.entry_name = f'{name[0]} data file'

        if name[-1] == 'xlsx':
            data_frame = pd.read_excel(mainfile)
        elif name[-1] == 'csv':
            data_frame = pd.read_csv(mainfile)
        else:
            return
        logger.info(f'Parsing {filename} with {data_frame.shape[0]} rows')
    
        for col in data_frame.columns:
            col_small = col.strip().casefold()
            data_frame.rename(columns={col: col_small}, inplace=True)

        if 'CatalyticReactionCollection' in name[-2]:
            self.extract_reaction_entries(data_frame, archive, logger)
            logger.info(
                f'''File {filename} matches the expected format for a reaction
                collection. Reaction entries are successfully extracted.'''
            )
            return
        elif 'CatalystSampleCollection' in name[-2]:
            try:
                data_frame = self.unify_columnnames(data_frame)
                samples = self.extract_sample_entries(data_frame, archive, logger)
                logger.info(
                    f'''File {filename} matches the expected format for a catalysis
                    collection. Sample entries successfully extracted.'''
                )        
                samples_references = []
                for sample in samples:
                    sample_ref = CompositeSystemReference(
                        reference = sample,
                    )
                    samples_references.append(sample_ref)
                archive.data.samples = samples_references
                return
            except Exception as e:
                logger.error(f'Error extracting sample entries: {e}')
        
        elif 'CatalysisCollection' in name[-2]:
            try:
                self.extract_reaction_entries(data_frame, archive, logger)
                logger.info(
                    f'''File {filename} matches the expected format for a catalysis
                    collection. Reaction entries successfully extracted. And sample 
                    entries will be extracted next'''
                )
            except Exception as e:
                logger.error(f'Error extracting reaction entries: {e}')
                return
            
            try:
                data_frame = self.unify_columnnames(data_frame)
                samples = self.extract_sample_entries(data_frame, archive, logger)
                logger.info(
                    f'''File {filename} matches the expected format for a catalysis
                    collection. Sample entries successfully extracted.'''
                )
                samples_references = []
                for sample in samples:
                    sample_ref = CompositeSystemReference(
                        reference = sample,
                    )
                    samples_references.append(sample_ref)
                archive.data.samples = samples_references

            except Exception as e:
                logger.error(f'Error extracting sample entries: {e}')
        
        return



import os.path

import pytest
from nomad.client import normalize_all, parse


def test_schema():
    test_file = os.path.join('tests', 'data', 'test_sample.archive.yaml')
    entry_archive = parse(test_file)[0]
    normalize_all(entry_archive)

    assert entry_archive.data.name == 'MoO3'
    assert entry_archive.results.material.elements == ['Mo', 'O']
    assert entry_archive.results.properties.catalytic.catalyst.catalyst_name == 'MoO3'
    assert (
        entry_archive.results.properties.catalytic.catalyst.preparation_method
        == 'spray-drying'
    )
    assert entry_archive.results.properties.catalytic.catalyst.catalyst_type == [
        'bulk catalyst',
        'oxide',
    ]
    assert (
        entry_archive.results.properties.catalytic.catalyst.characterization_methods
        == ['BET']
    )
    assert entry_archive.results.properties.catalytic.catalyst.surface_area.to(
        'm^2/g'
    ).magnitude == pytest.approx(3.27)

    test_file = os.path.join('tests', 'data', 'test_reaction.archive.yaml')
    entry_archive = parse(test_file)[0]
    normalize_all(entry_archive)

    assert entry_archive.data.name == 'methanol dehydrogenation 30'
    assert entry_archive.data.reaction_conditions.set_temperature.to(
        'K'
    ).magnitude == pytest.approx(473)



import os.path

from nomad.client import normalize_all, parse


def test_parser():
    test_file = os.path.join('tests', 'data', 'template_CatalystSampleCollection.xlsx')
    entry_archive = parse(test_file)[0]
    normalize_all(entry_archive)

    assert (
        entry_archive.metadata.entry_name
        == 'template_CatalystSampleCollection data file'
    )
    number_of_samples = 2
    assert len(entry_archive.data.samples) == number_of_samples

    test_file = os.path.join('tests', 'data', 'template_CatalyticReaction.xlsx')
    entry_archive = parse(test_file)[0]
    normalize_all(entry_archive)

    assert entry_archive.metadata.entry_name == 'template_CatalyticReaction data file'
    assert entry_archive.metadata.entry_type == 'RawFileData'



from nomad.config.models.plugins import SchemaPackageEntryPoint
from pydantic import Field


class CatalysisPackageEntryPoint(SchemaPackageEntryPoint):
    parameter: int = Field(0, description='Custom configuration parameter')

    def load(self):
        from nomad_catalysis.schema_packages.catalysis import m_package # noqa: PLC0415, I001

        return m_package


catalysis = CatalysisPackageEntryPoint(
    name='Catalysis',
    description='Catalysis Schema package defined using the new plugin mechanism.',
)
